"""
    Utitlity functions for vector(x,y,z)
    ----------------------------------------------------------

    file under vectogebra - python module for vector algebra

    ----------------------------------------------------------

    vectogebra/src/utilities.py

    -------------------------
    This module contains the utility operations for vasious calculation related to vectors.
    
    requires: vectogebra/src/vector.py to be imported
    
    depends on: 
    1. vectogebra/src/vector.py
    2. math module 
    -------------------------------------------------
    copyright: (C) 2022 Mohammad Maasir
    license: MIT License (see LICENSE in project's main directory)
    -------------------------------------------------
    date created: 9th of May, 2022 (3:09 AM)
    last modified: 9th of May, 2022
    -------------------------------------------------

    contributor(s): Mohammad Maasir

    -------------------------------------------------

    github: github.com/maasir554/vectogebra/vectogrbra/utilities.py

    ----------------------------------------------------------------

    report an issue : https://github.com/maasir554/vectogebra/issues
    
    ---------------------------------------------------------------------
    
    email: maasir554@gmail.com

    ---------------------------------------------------------------------

    

"""

# ------------------------------------------------------(result of a lot of research, trial and error)
# sys.path is where python search for modules.

import pathlib
import sys

currentDIR = pathlib.Path(__file__).resolve().parent
sys.path.append(str(currentDIR))
# ------------------------------------------------------  (added the src/vectogebra to the path for the test file to work)


# importlib is imported for making imports easier.


import importlib


# importing math for trigonometric and inverse trigonometric functions :-


import math


# importing vector class from vector.py


try:
    from .vector import vector as v
except:
   v = importlib.import_module('vector').vector
#    from vector import vector as v

# print(v)
# print(v(1,1,1)*v(2,2,2))

# NOTE: the type of vector will also be reffered as alias i.e. in this case, v only.
# example: a = v(1,2,3)
# print(type(a) == v) => This will return True. NOTE: do NOT use "v" or 'v' or vector.v ,
# just use alias as it is i.e.: v
# if we would have taken alias as vect then we would refer to it as vect only. and not "vect" or 'vect' or vector.vect
# if we did not had used alias, then we would refer to it as vector.vector
# Please don't get confused when working outside this module. there will be different conventions.

##########################################################

############# _____  ANGLE RELATED _____ ##################

##########################################################

# for angle between two vectors options : "radians"(default) or "degrees"
def angle(v1, v2, format="radians"):
    # python is bad at multipilcation of irrational numbers hence, we use magnitude_square()
    # instead of magnitude() (as magnitude_squared is defined separately, without use of square roots.)
    # else we  get errors like 5.99999999999999999 != 6 xD
    if (v1*v2)**2 != (v1.magnitude_squared)*(v2.magnitude_squared) and type(v1) == type(v2) == v:
        if format == "radians":
            return math.acos((v1 * v2) / (v1.magnitude * v2.magnitude))
        elif format == "degrees":
            return math.degrees(math.acos((v1 * v2) / (v1.magnitude * v2.magnitude)))
        else:
            raise ValueError("third argument must be either 'radians' or 'degrees'")
    elif (v1*v2)**2 == (v1.magnitude_squared)*(v2.magnitude_squared) and type(v1) == type(v2) == v:
        return 0.0
    else :
        raise TypeError("First two arguments must be vectors and optionally, third argument must be either 'radians' or 'degrees'")

# function to check if the tow vectors are perpendicular or not
def is_perpendicular(v1, v2):
    if angle(v1, v2, 'degrees') == 90.0:
        return True
    else:
        return False


# function to check if the tow vectors are parallel or not
def is_parallel(v1, v2):
    if angle(v1, v2, 'degrees') == 0.0:
        return True
    else:
        return False


# function to obtain unit vector from a given vector
def unit_vector(v):
    return (v / v.magnitude)


def direction(v):
    return (v / v.magnitude)


###########################################################

################_________ ALGEBRA _______###################

###########################################################

# addition of n vectors
def add(*args):
    v0 = v(0, 0, 0)
    for i in args:
        if type(i) == v:
            v0 = v0 + i
        elif type(i) != v:
            print(type(i))
            print(type(v))
            raise TypeError(type(v), type(i), "Arguments must be vectors")
        else:
            return 'Something went wrong :-('
    return (v0)


# function for dot product
def dot(v1, v2):
    return v1 * v2


# function for cross product
def cross(v1, v2):
    return v1 ^ v2


# magnitude of a vector or a number
def magnitude(v):
    if type(v) == v:
        return v.magnitude
    elif type(v) == int or type(v) == float:
        return (v ** 2) ** 0.5
    else:
        raise TypeError("Argument must be a vector or a number")


###########################################################

################_________ GEOMETRY _______##################

###########################################################

# function to find area of parallelogram formed by two vectors
def parallelogram_area(a, b):
    return a.magnitude * b.magnitude * math.sin(angle(a, b, 'radians'))


# function to find volume of parallelopiped formed by three vectors or box product
def box(a, b, c):
    return a * (b ^ c)


# function to check if three vectors are coplanar
def coplanar(a, b, c):
    if box(a, b, c) == 0:
        return (True)
    else:
        return (False)


# scalar component of vector a along the direction of vector b or opposite to direction of vector b(if the angle is
# obtuse)
def scalar_component_collinear(a, b):
    if type(a) == v and type(b) == v:
        return a * unit_vector(b)
    else:
        raise TypeError("Arguments must be vectors")


# vector component of vector a along the direction of vector b or opposite to direction of vector b(if the angle is
# obtuse)
def vector_component_collinear(a, b):
    return a * unit_vector(b) * unit_vector(b)


# scalar component perpendicular to direction of vector b
def scalar_component_perpendicular(a, b):
    return (a ^ b) / b.magnitude


# vector component perpendicular to direction of vector b
def vector_component_perpendicular(a, b):
    return scalar_component_perpendicular(a, b) * unit_vector(b ^ (a ^ b))  # Thumb rule logics***


###########################################################

#############________ EXPORT A VECTOR _______##############

###########################################################


# dictionary from a vector 
def vector_to_dict(v):
    return {'x': v.x, 'y': v.y, 'z': v.z}


# list from a vector (it will contain its coordinates/components)
def vector_to_list(v):
    return [v.x, v.y, v.z]  # returning a list of coordinates/components arranged in order of x,y,z


# tuple from a vector (it will contain its coordinates/components)
def vector_to_tuple(v):
    return v.x, v.y, v.z


###########################################################

############# CREATION OF VECTOR  #########################

###########################################################

# vector from a dictionary
def dict_to_vector(d):
    if type(d) == dict:
        return v(d['x'], d['y'], d['z'])
    else:
        raise TypeError("Argument must be a dictionary")


# vector from a list


def list_to_vector(l):
    if type(l) == list:

        return (v(l[0], l[1], l[2]))

    else:

        raise TypeError("Argument must be a list")


# vector from a tuple

def tuple_to_vector(t):
    if type(t) == tuple:
        return v(t[0], t[1], t[2])
    else:
        raise TypeError("Argument must be a tuple")


# 2d vector from polar coordinates


def polar_to_vector(r, theta):
    return v(r * math.cos(theta), r * math.sin(theta), 0)




"""
---------------------------------------------------------------------
vectogebra/test/test_attributes.py
---------------------------------------------------------------------
File under project "vectogebra"
---------------------------------------------------------------------
This file contains the test cases for the attributes of vector class.
---------------------------------------------------------------------
License: MIT License (see LICENSE in project's main directory)
copyright: (C) 2022 Mohammad Maasir
---------------------------------------------------------------------
- date created: 9th of May, 2022 (09:55 PM)
- last modified: 9th of May, 2022
---------------------------------------------------------------------
contributor(s): Mohammad Maasir
---------------------------------------------------------------------
github: github.com/maasir554/vectogebra/vectogrbra/vector.py
---------------------------------------------------------------------
report an issue at: github.com/maasir554/vectogebra/issues
---------------------------------------------------------------------
email: maasir554@gmail.com


"""

# NOTE: do NOT run this file directly.

# Instead, open console or terminal and change dirctory to root directory of this project : 
# $ cd vectogebra
# then, run the following command:
# python -m unittest test/test_attributes.py 
# OR
# python -m unittest test.test_attributes
# OR 
# python -m unittest discover 

## and you should see an OK message.
## if not, then you have some test case(s) that failed.

#------------------------------------------------------
import pathlib
import sys
currentDIR = pathlib.Path(__file__).resolve().parent
rootDIR = currentDIR.parent
vectogebraMAIN = rootDIR / 'src' /'vectogebra'
sys.path.append(str(vectogebraMAIN))
sys.path.append(str(rootDIR))
#------------------------------------------------------

import unittest

from src.vectogebra.vector import vector as vect
#from vector import vector as vect 
#do NOT delete the above comment.

v1 = vect(1,2,3)
v2 = vect(4,5,6)
v3 = vect(2,2,2)

class TestStringMethods(unittest.TestCase):

    ############################################
    ######## VECTOR CLASS ATTRIBUTES ###########
    ############################################

    def test_vector_components(self):
        self.assertEqual(v1.i, 1)
        self.assertEqual(v1.j, 2)
        self.assertEqual(v1.k, 3)
    

    def test_magnitude(self):
        self.assertEqual(v1.magnitude, 3.7416573867739413)
    
    def test_magnitude_squared(self):
        self.assertEqual(v1.magnitude_squared,14)

    def test_unit_vector(self):
        self.assertEqual(v1.unit_vector, vect(0.2672612419124244,0.5345224838248488,0.8017837257372732))
    
    
    


    # def test_subctraction(self):
    #     self.assertTrue('FOO'.isupper())
    #     self.assertFalse('Foo'.isupper())

    # def test_split(self):
    #     s = 'hello world'
    #     self.assertEqual(s.split(), ['hello', 'world'])
    #     # check that s.split fails when the separator is not a string
    #     with self.assertRaises(TypeError):
    #         s.split(2)

if __name__ == '__main__':
    unittest.main()



"""
---------------------------------------------------------------------
vectogebra/test/test_algebra.py
---------------------------------------------------------------------
File under project "vectogebra"
---------------------------------------------------------------------
This file contains the test cases for the attributes of vector class.
---------------------------------------------------------------------
License: MIT License (see LICENSE in project's main directory)
copyright: (C) 2022 Mohammad Maasir
---------------------------------------------------------------------
- date created: 9th of May, 2022 (11:35 PM)
- last modified: 9th of May, 2022
---------------------------------------------------------------------
contributor(s): Mohammad Maasir
---------------------------------------------------------------------
github: github.com/maasir554/vectogebra/vectogrbra/vector.py
---------------------------------------------------------------------
report an issue at: github.com/maasir554/vectogebra/issues
---------------------------------------------------------------------
email: maasir554@gmail.com


"""

# NOTE: do NOT run this file directly.

# Instead, open console or terminal and change dirctory to root directory of this project : 
# $ cd vectogebra
# then, run the following command:
# python -m unittest test/test_algebra.py 
# OR
# python -m unittest test.test_algebra
# OR 
# python -m unittest discover 

## and you should see an OK message.
## if not, then you have some test case(s) that failed.

#####################################################################

############## the following block is VERY important ################

#####################################################################
#------------------------------------------------------(result of a lot of research, trial and error)
import pathlib
import sys
currentDIR = pathlib.Path(__file__).resolve().parent
rootDIR = currentDIR.parent
vectogebraMAIN = rootDIR / 'src' /'vectogebra'
sys.path.append(str(vectogebraMAIN))
sys.path.append(str(rootDIR))
# os module can also be used here. but pathlib is easier to work with
#------------------------------------------------------  (added the src/vectogevra to the path for the test file to work)
import unittest

# from vector import vector as vect

from src.vectogebra.vector import vector as vect
# if you wish to run directly, you can use the following:(disable the above line first)
# from vector import vector as vect
v1 = vect(1,2,3)
v2 = vect(4,5,6)
v3 = vect(2,2,2)


class TestStringMethods(unittest.TestCase):
    
    # ########################################## #

    # ######### ALGEBRIC OPERATORS ############# #

    # ########################################## #

    def test_addition_operator(self):
        self.assertEqual(v1+v2, vect(5,7,9))
        self.assertEqual(v1+v2+v3, vect(7,9,11))

    def test_subctraction_operator(self):
        self.assertEqual(v1-v2, vect(-3,-3,-3))
        self.assertEqual(v1-v2-v3, vect(-5,-5,-5))

    def test_dot_operator(self):
        self.assertEqual(v1*v2, 32)
        self.assertEqual(v1*v2*v3, 32*v3)

    def test_cross_operator(self):
        self.assertEqual(v1^v2, vect(-3,6,-3))
        self.assertEqual(v1^v2^v3, vect(18,0,-18))  #same as v1^(v2^v3)
    
    def test_scalar_multiplication(self):
        self.assertEqual(v1*2, vect(2,4,6))
        self.assertEqual(2*v1, vect(2,4,6))

    def test_scalar_division(self):
        self.assertEqual(v1/2, vect(0.5,1,1.5))


if __name__ == '__main__':
    unittest.main()



"""
---------------------------------------------------------------------
vectogebra/test/test_algebra.py
---------------------------------------------------------------------
File under project "vectogebra"
---------------------------------------------------------------------
This file contains the test cases for the attributes of vector class.
---------------------------------------------------------------------
License: MIT License (see LICENSE in project's main directory)
copyright: (C) 2022 Mohammad Maasir
---------------------------------------------------------------------
- date created: 9th of May, 2022 (11:35 PM)
- last modified: 9th of May, 2022
---------------------------------------------------------------------
contributor(s): Mohammad Maasir
---------------------------------------------------------------------
github: github.com/maasir554/vectogebra/vectogrbra/vector.py
---------------------------------------------------------------------
report an issue at: github.com/maasir554/vectogebra/issues
---------------------------------------------------------------------
email: maasir554@gmail.com


"""

# NOTE: do NOT run this file directly.

# Instead, open console or terminal and change dirctory to root directory of this project : 
# $ cd vectogebra
# then, run the following command:
# python -m unittest test/test_algebra.py 
# OR
# python -m unittest test.test_algebra
# OR 
# python -m unittest discover 

## and you should see an OK message.
## if not, then you have some test case(s) that failed.

#------------------------------------------------------
import pathlib
import sys
currentDIR = pathlib.Path(__file__).resolve().parent
rootDIR = currentDIR.parent
vectogebraMAIN = rootDIR / 'src' /'vectogebra'
sys.path.append(str(vectogebraMAIN))
sys.path.append(str(rootDIR))
#------------------------------------------------------


import unittest


#from vector import vector as vect
#import utilities as vut
""" 
    enable the above lines and comment two 
    the lines below
    to run this file directly.
"""

import math

from src.vectogebra.vector import vector as vect

from src.vectogebra import utilities as vut

v1 = vect(1,2,3)
v2 = vect(4,5,6)
v3 = vect(2,2,2)
v4= vect(20,20,20)

class TestStringMethods(unittest.TestCase):
    
    ############################################

    #### UTILITY FUNCTIONS FOR CLASS VECTOR ####

    ############################################

    def test_add_function(self):
        self.assertEqual(vut.add(v1,v2), vect(5,7,9))
        self.assertEqual(vut.add(v1,v2,v3), vect(7,9,11)) # Any number of vectors can be added.
    
    def test_angle_function(self):
        self.assertEqual(vut.angle(v1,v2), 0.2257261285527342)
        self.assertEqual(vut.angle(v3,v4), 0)
        self.assertEqual(vut.angle(vect(1,0,0),vect(0,1,0)), math.pi/2)



if __name__ == '__main__':
    unittest.main()


"""
    file under vectogebra - python module for vector algebra
    
    vectogrbra/__init__.py
    -------------------------
    This module contains the 'vector' class.
    -------------------------------------------------
    copyright: (C) 2022 Mohammad Maasir
    license: MIT License (see LICENSE in project's main directory)
    -------------------------------------------------
    date created: 9th of May, 2022 (3:09 AM)
    last modified: 9th of May, 2022
    -------------------------------------------------
    contributor(s): Mohammad Maasir
    -------------------------------------------------
    github: github.com/maasir554

"""

# importing vector class from vector.py
# the dot is necessary to avoid name conflict. 
# from .vector import vector

# exception handling is best solution to any problem
# here in
try:
    from .vector import vector  # this import is for the build version
except :
    pass        # simply ignore the error and the module will be imported 
# pass will simply ignore the import command here and test files will import the module by
# the local command written in them


"""
    to use vector object class, in your file, write: import vectogerbra.vector as v
    then you can use the vector class in your file
    example: v1 = v(1,2,3)

    to import the vector class in your file, you can use the following:
    ## import vectogerbra.vector as v [or any other alias name you want]
    ## from vectogerbra import vector as v [or any other alias name you want]
"""

# importing utility functions from utilities.py

# from .utilities import *




"""
    vector(x,y,z)
    ----------------------------------------------------------------

    file under vectogebra - python module for vector algebra

    ----------------------------------------------------------------

    vectogebra/src/vector.py

    ----------------------------------------------------------------

    This module contains the 'vector' class.

    ----------------------------------------------------------------
    copyright: (C) 2022 Mohammad Maasir
    license: MIT License (see LICENSE in project's main directory)
    ----------------------------------------------------------------
    date created: 8th of May, 2022 (11:45 PM)
    last modified: 8th of May, 2022
    ----------------------------------------------------------------

    contributor(s): Mohammad Maasir

    ----------------------------------------------------------------

    github: github.com/maasir554/vectogebra/vectogrbra/vector.py
    
    ----------------------------------------------------------------

    report an issue : https://github.com/maasir554/vectogebra/issues
    
    ---------------------------------------------------------------------
    
    email: maasir554@gmail.com

"""
import math


class vector:
    
    #Constructor

    def __init__(self, i, j, k):
        #Rectangular Components in i,j,k
        self.i = i
        self.j = j
        self.k = k
        #Rectangular Components in x,y,z
        self.x = i
        self.y = j
        self.z = k
        #Magnitude
        self.magnitude = self.__magnitude()  #refer to __magnitude() below
        self.modulus = self.magnitude
        self.mod = self.magnitude   
        self.mag = self.magnitude 
        #Magnitude squared
        self.magnitude_squared = self.__magnitude_squared() #refer to __magnitude_squared() below
        self.modulus_squared = self.magnitude_squared
        self.mod_squared = self.magnitude_squared
        self.mag_squared = self.magnitude_squared
        #Unit Vector [Tuple] along the vector (parallel)
        # NOTE !WARNING! the unit_vector function returns a TUPLE (x,y,z) and NOT a vector
        self.unit_vector = (self.__unit_vector())


        self.type = self.__type__()

    #String representation of the vector object :

    def __type__(self):
        return("vector")

    def __str__(self,format="default"):
        if format == "csv" or format == "tuple" or format == "default":
            return("vector({}, {}, {})".format(self.i, self.j, self.k))
        elif format == "componential" or format =="components" or format == "comp":
            return("({}i + {}j + {}k)".format(self.x, self.y, self.z))
        else:
            return("vector({}, {}, {})".format(self.x, self.y, self.z))

    def to_str(self,format="default") :
        return(self.__str__(format))

    def __repr__(self):
        return(self.__str__('default'))     # important for command line usage (when object is called without print())

    
    # Magnitude calculator

    def __magnitude(self):
        return math.sqrt((self.i)**2 + (self.j)**2 + (self.k)**2)

    # Magnitude Squared calculator

    def __magnitude_squared(self):
        return(self.i**2 + self.j**2 + self.k**2)

    # Unit vector calculator (will return simple Tuple only) the result will be a tuple of 3 numbers
    #! WARNING! This function will NOT return a vector object !

    def __unit_vector(self):
        if self.magnitude !=0:
            return(self.i/self.magnitude, self.j/self.magnitude, self.k/self.magnitude)
        else:
            return("Zero vector do not have a direction")
   
    # function to return unit_vector in different (string) formats 
    def unit_vector(self, format="default"):
        if format == "csv" or format == "tuple" or format == "default":
            return(self.unit_vector)
        elif format == "componential" or format =="components" or format == "comp":
            return("({}i + {}j + {}k)".format(self.x, self.y, self.z))

    # ALGEBRA : addition of vectors :-
    def __add__(self, other):
        return(vector(self.i + other.i, self.j + other.j, self.k + other.k))
    
    # ALGEBRA : DOT product and multiplication of vectors with scalars:-
    def __mul__(self, other):
        if type(other) == int or type(other) == float:
            return(vector(self.i * other, self.j * other, self.k * other))
        elif type(other) == vector:
            return(self.i * other.i + self.j * other.j + self.k * other.k)
        else:
            return("Error: Invalid input")
    #if salar number is wrir=tten in the beging, we uses __rmul__ overloader [Magic Method]
    def __rmul__(self, other):
        if type(other) == int or type(other) == float:
            return(vector(self.i * other, self.j * other, self.k * other))
        elif type(other) == vector:
            return(self.i * other.i + self.j * other.j + self.k * other.k)
        else:
            return("Error: Invalid input")

    # ALGEBRA : subtraction of vectors :-
    def __sub__(self, other):
        return(vector(self.i - other.i, self.j - other.j, self.k - other.k))

    # ALGEBRA : cross product of vectors :-
    # NOTE : ^(XOR operator) is the cross product operator in this case 
    def __xor__(self, other):
        if type(other) ==vector:
            return(vector(self.j * other.k - self.k * other.j, self.k * other.i - self.i * other.k, self.i * other.j - self.j * other.i))
        else:
            suggest = "if you have any suggestion, please suggest @ https://github.com/maasir554/vectogebra/issues"
            return("Error: Please note that the cross product is only defined for vectors. \nIn vectogebra, for any two vectors a and b, (a^b) is same as a cross b.\n" + suggest)
    
    # ALGEBRA : division by scalar
    def __truediv__(self, other):
        if type(other) == int or type(other) == float:
            return(vector(self.i / other, self.j / other, self.k / other))  
        else:
            suggest = "if you have any suggestion, please suggest @ https://github.com/maasir554/vectogebra/issues"
            return("Error: Please note that division of vector by a vector is not possible. \n" + suggest)
    # ALGEBRA : raising vectors to a integral power [Inspired by JEE queations] 
    def __pow__(self, other):
        if type(other) == int  and other == 0:
            return("Error! : power must be an integer >= 1")
        elif type(other) == int and other>=1 and other%2 == 0:
            return(self.magnitude**(other))
        elif type(other) == int and other>=3 and other%2 == 1:
            return((self.magnitude**(other-1))*self)
    # Making the class iterable to perform addition of multiple vectors simultainously e.g. a+b+c OR a-b+c OR a+b-c , etc.
    def __iter__(self):
        pass

    #############################################################################################

    # LOGICAL OPERATORS : equality of two vectors :-
    def __eq__(self, other):
        if type(other) == vector:
            return(self.i == other.i and self.j == other.j and self.k == other.k) #returns true if all components are equal
        else:
            return("Error: Vectors can only be equated among themselves")   #returns false if one or both of the input(s) is/are not (a) vector(s)
    # LOGICAL OPERATORS : inequality of two vectors :-
    # NOTE : != is the inequality operator in this case
    def __ne__(self, other):
        if type(other) == vector:
            return(self.i != other.i or self.j != other.j or self.k != other.k) #returns true if one or more of the components are not equal
        else:
            return("Error: Vectors can only be compared among themselves")
    # LOGICAL OPERATORS : greater than of two vectors :-
    def __gt__(self, other):
        if type(other) == vector:
            return(self.magnitude > other.magnitude)    # comparing magnitude makes sense instead of comparing components
        else:
            return("Error: Vectors can only be compared among themselves")
    # LOGICAL OPERATORS : less than of two vectors :-
    def __lt__(self, other):
        if type(other) == vector:
            return(self.magnitude < other.magnitude)    # comparing magnitude makes sense instead of comparing components
        else:
            return("Error: Vectors can only be compared among themselves")
    # LOGICAL OPERATORS : greater than or equal to of two vectors :-
    def __ge__(self, other):
        if type(other) == vector:
            return(self.magnitude >= other.magnitude)
        else:
            return("Error: Vectors can only be compared among themselves")
    # LOGICAL OPERATORS : less than or equal to of two vectors :-
    def __le__(self, other):
        if type(other) == vector:
            return(self.magnitude <= other.magnitude)
        else:
            return("Error: Vectors can only be compared among themselves")

    ########### More Dunder methods / Magic methods to be added #############




# i will write i cap as i_